import unittest
import json
from lazperf import Decompressor, Compressor, buildNumpyDescription, buildGreyhoundDescription
import numpy as np
import struct

schema = [{u'type': u'floating', u'name': u'X', u'size': 8}, {u'type': u'floating', u'name': u'Y', u'size': 8}, {u'type': u'floating', u'name': u'Z', u'size': 8}, {u'type': u'unsigned', u'name': u'Origin', u'size': 8}, {u'type': u'unsigned', u'name': u'Intensity', u'size': 2}, {u'type': u'unsigned', u'name': u'ReturnNumber', u'size': 1}, {u'type': u'unsigned', u'name': u'NumberOfReturns', u'size': 1}, {u'type': u'unsigned', u'name': u'ScanDirectionFlag', u'size': 1}, {u'type': u'unsigned', u'name': u'EdgeOfFlightLine', u'size': 1}, {u'type': u'unsigned', u'name': u'Classification', u'size': 1}, {u'type': u'floating', u'name': u'ScanAngleRank', u'size': 4}, {u'type': u'unsigned', u'name': u'UserData', u'size': 1}, {u'type': u'unsigned', u'name': u'PointSourceId', u'size': 2}, {u'type': u'floating', u'name': u'GpsTime', u'size': 8}]

len_compressed = 25691
len_uncompressed = 54112
expected_point_count = 1002

class TestLazPerf(unittest.TestCase):

    def test_decompressor(self):
        s = json.dumps(schema)

        with open('test/compressed.bin', 'rb') as f:
            data = f.read()

        with open('test/uncompressed.bin','rb') as f:
            original = f.read()

        self.assertEqual(len(data),
                         len_compressed,
                         "compressed file length is correct")
        self.assertEqual(len(original),
                         len_uncompressed,
                         "uncompressed file length is correct")

        # last four bytes are the point count
        compressed_point_count = struct.unpack('<L',data[-4:])[0]
        uncompressed_point_count = struct.unpack('<L',original[-4:])[0]

        self.assertEqual(compressed_point_count,
                         uncompressed_point_count,
                         "compressed point count matches expected")
        self.assertEqual(uncompressed_point_count,
                         expected_point_count,
                         "uncompressed point count matches expected")

        arr = np.frombuffer(data, dtype = np.uint8)
        dtype = buildNumpyDescription(json.loads(s))
        self.assertEqual(dtype.itemsize, 54)

        d = Decompressor(arr, s)
        decompressed = d.decompress(compressed_point_count)
        uncompressed = np.frombuffer(original[0:-4], dtype = dtype)

        self.assertEqual(uncompressed.shape[0], expected_point_count)
        self.assertEqual(decompressed.shape[0], expected_point_count)
        for i in range(len(uncompressed)):
           self.assertEqual(uncompressed[i], decompressed[i])

    def test_compressor(self):
        s = json.dumps(schema)

        with open('test/compressed.bin', 'rb') as f:
            data = f.read()

        with open('test/uncompressed.bin','rb') as f:
            original = f.read()

        self.assertEqual(len(data),
                         len_compressed,
                         "compressed file length is correct")
        self.assertEqual(len(original),
                         len_uncompressed,
                         "uncompressed file length is correct")

        # last four bytes are the point count
        compressed_point_count = struct.unpack('<L',data[-4:])[0]
        uncompressed_point_count = struct.unpack('<L',original[-4:])[0]

        self.assertEqual(compressed_point_count,
                         uncompressed_point_count,
                         "compressed point count matches expected")
        self.assertEqual(uncompressed_point_count,
                         expected_point_count,
                         "uncompressed point count matches expected")

        dtype = buildNumpyDescription(json.loads(s))

        uncompressed = np.frombuffer(original[0:-4], dtype = dtype)
        self.assertEqual(uncompressed.shape[0], expected_point_count)

        point_data = np.frombuffer(original[:-4], dtype = dtype)
        empty = np.zeros(uncompressed_point_count, dtype = np.uint8)

        c = Compressor(s)

        compressed = c.compress(point_data)
        original_compressed = np.frombuffer(data[0:-4], dtype = np.uint8)

        self.assertEqual(len(original_compressed), len_compressed - 4)
        for i in range(len(compressed)):
            self.assertEqual(compressed[i], original_compressed[i])

    def test_full_loop(self):
        s = json.dumps(schema)

        with open('test/uncompressed.bin','rb') as f:
            original = f.read()

        dtype = buildNumpyDescription(json.loads(s))
        uncompressed = np.frombuffer(original[0:-4], dtype = dtype)

        c = Compressor(s)
        compressed = c.compress(uncompressed)

        d = Decompressor(compressed, s)
        decompressed = d.decompress(expected_point_count)
        self.assertEqual(len(decompressed), len(uncompressed))
        for i in range(len(decompressed)):
            self.assertEqual(decompressed[i], uncompressed[i])

        # confirm we can build from dtypes instead of json descriptions
        c2 = Compressor(dtype)
        d2 = Decompressor(compressed, dtype)



def test_suite():
    return unittest.TestSuite(
        [TestLazPerf])

if __name__ == '__main__':
    unittest.main()




import unittest
import numpy as np
import struct
from lazperf import VLRDecompressor, VLRCompressor, RecordSchema, LazVLR



las_header_size = 227
vlr_header_size = 54
offset_to_laszip_vlr_data = las_header_size + vlr_header_size
laszip_vlr_data_size = 52
gt_offset_to_point_data = offset_to_laszip_vlr_data + laszip_vlr_data_size
sizeof_chunk_table_offset = 8
point_count = 1065

point_dtype = np.dtype([('X', '<i4'), ('Y', '<i4'), ('Z', '<i4'), ('intensity', '<u2'), ('bit_fields', 'u1'), ('raw_classification', 'u1'), ('scan_angle_rank', 'i1'), ('user_data', 'u1'), ('point_source_id', '<u2'), ('gps_time', '<f8'), ('red', '<u2'), ('green', '<u2'), ('blue', '<u2')])



class TestVLRDecompress(unittest.TestCase):
    def test_decompression(self):

        with open('test/simple_points_uncompressed.bin', mode='rb') as fin:
            groud_truth  = fin.read()

        with open('test/simple.laz', mode="rb") as fin:
            raw_data = fin.read()

        laszip_vlr_data = raw_data[
            offset_to_laszip_vlr_data: offset_to_laszip_vlr_data + laszip_vlr_data_size ]

        raw_points = raw_data[gt_offset_to_point_data + sizeof_chunk_table_offset:]

        laszip_vlr_data = np.frombuffer(laszip_vlr_data, dtype=np.uint8)
        compressed_points = np.frombuffer(raw_points, dtype=np.uint8)

        decompressor = VLRDecompressor(compressed_points, laszip_vlr_data)
        points_decompressed = decompressor.decompress_points(point_count)

        points_decompressed = np.frombuffer(points_decompressed, dtype=point_dtype)
        points_groud_truth = np.frombuffer(groud_truth, dtype=point_dtype)

        self.assertTrue(np.all(points_groud_truth == points_decompressed))

class TestVLRCompress(unittest.TestCase):


    def test_compression(self):
        with open('test/simple_points_uncompressed.bin', mode='rb') as fin:
            points_to_compress = np.frombuffer(fin.read(), dtype=np.uint8)

        with open('test/simple.laz', mode="rb") as fin:
            ground_truth = fin.read()

        laszip_vlr_data = ground_truth[
            offset_to_laszip_vlr_data: offset_to_laszip_vlr_data + laszip_vlr_data_size ]

        gt_vlr_data = np.frombuffer(laszip_vlr_data, dtype=np.uint8)
        gt_points_compressed = np.frombuffer(ground_truth[gt_offset_to_point_data + sizeof_chunk_table_offset:], dtype=np.uint8)

        rs = RecordSchema()
        rs.add_gps_time()
        rs.add_rgb()

        vlr = LazVLR(rs)
        vlr_data = vlr.data()

        self.assertEqual(vlr.data_size(), laszip_vlr_data_size)
        self.assertTrue(np.all(vlr_data == gt_vlr_data))

        offset_to_data = offset_to_laszip_vlr_data + vlr.data_size()
        self.assertEqual(offset_to_data, gt_offset_to_point_data)

        compressor = VLRCompressor(rs, offset_to_data)
        points_compressed = compressor.compress(points_to_compress)

        chunk_table_offset = struct.unpack('<Q', points_compressed[:sizeof_chunk_table_offset].tobytes())[0]
        gt_chunk_table_offset = struct.unpack('<Q', ground_truth[gt_offset_to_point_data:gt_offset_to_point_data + sizeof_chunk_table_offset])[0]
        self.assertEqual(chunk_table_offset, gt_chunk_table_offset)

        # This is the chunk table offset relative to the point data
        # without las header and vlrs
        relative_chunk_table_offset = chunk_table_offset - offset_to_data
        chunk_table = points_compressed[relative_chunk_table_offset:].tobytes()
        gt_chunk_table = ground_truth[gt_chunk_table_offset:]
        self.assertEqual(chunk_table, gt_chunk_table)

        points_compressed = points_compressed[sizeof_chunk_table_offset:].tobytes()
        gt_points_compressed = ground_truth[gt_offset_to_point_data + sizeof_chunk_table_offset:]
        self.assertEqual(points_compressed, gt_points_compressed)

    def test_raises(self):
        with open('test/simple_points_uncompressed.bin', mode='rb') as fin:
            point_buffer = fin.read()

        rs = RecordSchema()
        rs.add_gps_time()
        rs.add_rgb()

        vlr = LazVLR(rs)
        vlr_data = vlr.data()

        # Cut off point data to check if exception is properly raised
        points_to_compress = np.frombuffer(point_buffer[:-12], dtype=np.uint8)
        offset_to_data = offset_to_laszip_vlr_data + vlr.data_size()
        compressor = VLRCompressor(rs, offset_to_data)

        with self.assertRaises(ValueError):
            points_compressed = compressor.compress(points_to_compress)

if __name__ == '__main__':
    unittest.main()



__version__='1.2.1'
from .pylazperfapi import PyDecompressor as Decompressor
from .pylazperfapi import PyCompressor as Compressor
from .pylazperfapi import PyVLRDecompressor as VLRDecompressor
from .pylazperfapi import PyVLRCompressor as VLRCompressor
from .pylazperfapi import PyRecordSchema as RecordSchema
from .pylazperfapi import PyLazVlr as LazVLR
from .pylazperfapi import buildNumpyDescription
from .pylazperfapi import buildGreyhoundDescription




# distutils: language = c++

from libcpp.vector cimport vector
from libcpp.string cimport string
from libc.stdint cimport uint8_t, int32_t, uint64_t
from cpython.version cimport PY_MAJOR_VERSION
from cpython.array cimport array, clone

import json as jsonlib
import numpy as np

cimport lazperf
cimport numpy as np
np.import_array()

def get_lazperf_type(size, t):
    if t == 'floating':
        if size == 8:
            return lazperf.Double
        else:
            return lazperf.Float
    if t == 'unsigned':
        if size == 8:
            return lazperf.Unsigned64
        elif size == 4:
            return lazperf.Unsigned32
        elif size == 2:
            return lazperf.Unsigned16
        elif size == 1:
            return lazperf.Unsigned8
        else:
            raise Exception("Unexpected type size '%s' for unsigned type" % size)
    if t == 'signed':
        if size == 8:
            return lazperf.Signed64
        elif size == 4:
            return lazperf.Signed32
        elif size == 2:
            return lazperf.Signed16
        elif size == 1:
            return lazperf.Signed8
        else:
            raise Exception("Unexpected type size '%s' for signed type" % size)

def buildNumpyDescription(schema):
    """Given a Greyhound schema, convert it into a numpy dtype description
    http://docs.scipy.org/doc/numpy/reference/generated/numpy.dtype.html
    """
    formats = []
    names = []

    try:
        schema[0]['type']
    except:
        schema = jsonlib.loads(schema)

    for s in schema:
        t = s['type']
        if t == 'floating':
            t = 'f'
        elif t == 'unsigned':
            t = 'u'
        else:
            t = 'i'

        f = '%s%d' % (t, int(s['size']))
        names.append(s['name'])
        formats.append(f)
    return np.dtype({'names': names, 'formats': formats})

def buildGreyhoundDescription(dtype):
    """Given a numpy dtype, return a Greyhound schema"""
    output = []
    for t in dtype.descr:
        name = t[0]
        dt = dtype.fields[name]
        size = dt[0].itemsize
        tname = dt[0].name

        entry = {}
        if 'float' in tname:
            entry['type'] = 'floating'
        elif 'uint' in tname:
            entry['type'] = 'unsigned'
        else:
            entry['type'] = 'signed'

        entry['size'] = size
        entry['name'] = name
        output.append(entry)
    return output


cdef class PyCompressor:
    """ Class to compress points in the laz format using a json schema or numpy dtype
    to describe the point format
    """
    cdef lazperf.Compressor *thisptr      # hold a c++ instance which we're wrapping
    cdef public str jsondata
    cdef vector[uint8_t] *v

    def __init__(self, object schema):
        """
        schema: numpy dtype or json string of the point schema
        """
        self.v = new vector[uint8_t]()
        self.thisptr = new lazperf.Compressor(self.v[0])

        try:
            self.jsondata = jsonlib.dumps(buildGreyhoundDescription(schema))
        except AttributeError:
            self.jsondata = schema
        self.add_dimensions(self.jsondata)

    def compress(self, np.ndarray arr not None):
        """ Compresses points and return the result as a numpy array of bytes
        """

        cdef np.ndarray[uint8_t, ndim=1, mode="c"] view
        view = arr.view(dtype=np.uint8)

        point_count = self.thisptr.compress(view.data, view.shape[0])
        self.done()
        return self.get_data()

    cdef get_data(self):
        cdef const vector[uint8_t]* v = self.thisptr.data()
        cdef np.ndarray[uint8_t, ndim=1, mode="c"] arr = np.zeros(v.size(), dtype=np.uint8)

        self.thisptr.copy_data_to(<uint8_t*> arr.data)
        return arr

    def done(self):
        self.thisptr.done()

    def add_dimensions(self, jsondata):

        data = None
        try:
            jsondata[0]['type']
            data = jsondata
        except:
            data = jsonlib.loads(jsondata)

        for dim in data:
            t = get_lazperf_type(dim['size'], dim['type'])
            self.thisptr.add_dimension(t)

    def __dealloc__(self):
        del self.v
        del self.thisptr


cdef class PyDecompressor:
    """ Class to decompress laz points using a json schema/numpy dtype to
    describe the point format
    """
    cdef lazperf.Decompressor *thisptr      # hold a c++ instance which we're wrapping
    cdef public str jsondata

    def __init__(self, np.ndarray[uint8_t, ndim=1, mode="c"] compressed_points not None, object schema):
        """
        compressed_points: buffer of compressed_points
        schema: numpy dtype or json string of the point schema
        """
        try:
            self.jsondata = jsonlib.dumps(buildGreyhoundDescription(schema))
        except AttributeError:
            self.jsondata = schema

        self.thisptr = new lazperf.Decompressor(
            <const uint8_t*>compressed_points.data, compressed_points.shape[0])
        self.add_dimensions(self.jsondata)

    def decompress(self, size_t num_points):
        """ decompress points

        returns the numpy structured array of the decompressed points
        """
        cdef size_t point_size = self.thisptr.getPointSize()
        cdef np.ndarray[uint8_t, ndim=1, mode="c"] out = np.zeros(num_points * point_size, np.uint8)

        point_count = self.thisptr.decompress(out.data, out.shape[0])
        return out.view(dtype=buildNumpyDescription(self.jsondata))

    def add_dimensions(self, jsondata):
        data = None
        try:
            jsondata[0]['type']
            data = jsondata
        except:
            data = jsonlib.loads(jsondata)

        for dim in data:
            t = get_lazperf_type(dim['size'], dim['type'])
            self.thisptr.add_dimension(t)


    def __dealloc__(self):
        del self.thisptr

cdef class PyRecordSchema:
    """ This class is used to represent a LAS record schema
    This RecordSchema is nessecary for the LazVlr to be able to compress
    points meant to be written in a LAZ file.
    """
    cdef lazperf.record_schema schema

    def __init__(self):
        self.schema.push(lazperf.POINT10)

    def add_gps_time(self):
        self.schema.push(lazperf.GPSTIME)

    def add_rgb(self):
        self.schema.push(lazperf.RGB12)

cdef class PyLazVlr:
    """ Wraps a Lazperf's LazVlr class.
    This class is meant to give access to the Laszip's vlr raw record_data
    to allow writers to write LAZ files with its corresponding laszip vlr.
    """
    cdef lazperf.laz_vlr vlr
    cdef public PyRecordSchema schema

    def __init__(self, PyRecordSchema schema):
        self.schema = schema
        self.vlr = lazperf.laz_vlr.from_schema(schema.schema)

    def data(self):
        """ returns the laszip vlr record_data as a numpy array of bytes
        to be written in the VLR section of a LAZ compressed file.
        """
        cdef size_t vlr_size = self.vlr.size()
        cdef np.ndarray[uint8_t, ndim=1, mode="c"] arr = np.ndarray(vlr_size, dtype=np.uint8)

        self.vlr.extract(arr.data)
        return arr

    def data_size(self):
        """ Returns the number of bytes in the lazvlr record_data
        """
        return self.vlr.size()



cdef class PyVLRDecompressor:
    """ Class to decompress laz points stored in a .laz file using the
    Laszip vlr's record_data
    """
    cdef lazperf.VlrDecompressor *thisptr      # hold a c++ instance which we're wrapping

    def __init__(
            self,
            np.ndarray[uint8_t, ndim=1, mode="c"] compressed_points not None,
            np.ndarray[uint8_t, ndim=1, mode="c"] vlr not None
        ):
        """
        compressed_points: buffer of points to be decompressed
        vlr: laszip vlr's record_data as an array of bytes
        """
        cdef const uint8_t *p_compressed =  <const uint8_t*> compressed_points.data
        self.thisptr = new lazperf.VlrDecompressor(
            p_compressed, compressed_points.shape[0], vlr.data)

    def decompress_points(self, size_t point_count):
        """ decompress the points

        returns the decompressed data as an array of bytes
        """
        cdef size_t point_size = self.thisptr.getPointSize()
        cdef np.ndarray[uint8_t, ndim=1, mode="c"] point_out = np.zeros(point_size, dtype=np.uint8)
        cdef np.ndarray[uint8_t, ndim=1, mode="c"] points_uncompressed = np.zeros(point_count * point_size, dtype=np.uint8)
        cdef size_t i = 0
        cdef size_t begin = 0
        cdef size_t end = 0

        # Cython's memory views are needed to get the true C speed when slicing
        cdef uint8_t [:] points_view  = points_uncompressed
        cdef uint8_t [:] out_view = point_out

        for _ in range(point_count):
            self.thisptr.decompress(point_out.data)
            end = begin + point_size
            points_view[begin:end] = out_view
            begin = end

        return points_uncompressed

    def __dealloc__(self):
        del self.thisptr

cdef class PyVLRCompressor:
    """ Class to compress las points into laz format with the record schema
    from a laszip vlr, this is meant to be used by LAZ file writers
    """
    cdef lazperf.VlrCompressor *thisptr;

    def __init__(self, PyRecordSchema py_record_schema, uint64_t offset):
        """
        py_record_schema: The schema of the point format
        offset: offset to the point data (same as the las header field).
        This is needed because the first 8 bytes of the compressed points is an offset to the
        chunk table relative to the start of Las file. (Or you could pass in offset=0 and modify the
        8 bytes yourself)
        """
        self.thisptr = new lazperf.VlrCompressor(py_record_schema.schema, offset)


    def compress(self, np.ndarray arr):
        """ Returns the compressed points as a numpy array of bytes
        """
        cdef np.ndarray[char, ndim=1, mode="c"] view
        view = arr.view(np.uint8)

        cdef char *ptr = arr.data
        cdef size_t point_size = self.thisptr.getPointSize()
        cdef size_t num_bytes = arr.shape[0]
        cdef size_t point_count = num_bytes / point_size
        cdef double float_count = <double> num_bytes / point_size

        if arr.shape[0] % point_size != 0:
            raise ValueError("The number of bytes ({}) is not divisible by the point size ({})"
            " it gives {} points".format(num_bytes, point_size, float_count))

        for i in range(point_count):
            self.thisptr.compress(ptr)
            ptr += point_size

        self.thisptr.done()
        return self.get_data()

    cdef get_data(self):
        cdef const vector[uint8_t]* v = self.thisptr.data()
        cdef np.ndarray[uint8_t, ndim=1, mode="c"] arr = np.ndarray(v.size(), dtype=np.uint8)
        self.thisptr.copy_data_to(<uint8_t*>arr.data)
        return arr

    def __dealloc__(self):
        del self.thisptr




#!/usr/bin/env python

# Stolen from Shapely's setup.py
#

import logging
import os
import platform
import sys
import numpy
from distutils.version import StrictVersion

USE_CYTHON = False
try:

    from Cython.Build import cythonize
    import cython
    if not StrictVersion(cython.__version__) < StrictVersion('0.24'):
        USE_CYTHON = True

except ImportError:
    pass

ext = '.pyx' if USE_CYTHON else '.cpp'

from setuptools import setup


logging.basicConfig()
log = logging.getLogger(__file__)

# python -W all setup.py ...
if 'all' in sys.warnoptions:
    log.level = logging.DEBUG


# Get the version from the lazperf module
module_version = None
with open('lazperf/__init__.py', 'r') as fp:
    for line in fp:
        if line.startswith("__version__"):

            module_version = StrictVersion(
                line.split("=")[1].strip().strip("\"'"))
            break

if not module_version:
    raise ValueError("Could not determine lazperf's version")

# Handle UTF-8 encoding of certain text files.
open_kwds = {}
if sys.version_info >= (3,):
    open_kwds['encoding'] = 'utf-8'

with open('VERSION.txt', 'w', **open_kwds) as fp:
    fp.write(str(module_version))

with open('README.rst', 'r', **open_kwds) as fp:
    readme = fp.read()

with open('CHANGES.rst', 'r', **open_kwds) as fp:
    changes = fp.read()

long_description = readme + '\n\n' +  changes

include_dirs = ['.']
library_dirs = []
libraries = []
extra_link_args = []

from setuptools.extension import Extension as DistutilsExtension

include_dirs.append(numpy.get_include())
extra_compile_args = ['-std=c++11',]

DEBUG=True
if DEBUG:
    extra_compile_args += ['-g','-O0']

sources=['lazperf/pylazperfapi'+ext,"lazperf/PyLazperf.cpp",  "lazperf/PyVlrCompressor.cpp"]
extensions = [DistutilsExtension("lazperf.pylazperfapi",
                                   sources,
                                   include_dirs=include_dirs,
                                   library_dirs=library_dirs,
                                   extra_compile_args=extra_compile_args,
                                   extra_link_args=extra_link_args,
                                   language="c++",)]
if USE_CYTHON and "clean" not in sys.argv:
    from Cython.Build import cythonize
    extensions= cythonize(extensions, language="c++")

setup_args = dict(
    name                = 'lazperf',
    version             = str(module_version),
    install_requires    = ['numpy >=1.11'],
    description         = 'Point cloud data compression',
    license             = 'LGPL',
    keywords            = 'point cloud compression',
    author              = 'Howard Butler',
    author_email        = 'howard@hobu.co',
    maintainer          = 'Howard Butler',
    maintainer_email    = 'howard@hobu.co',
    url                 = 'https://github.com/hobu/laz-perf',
    long_description    = long_description,
    test_suite          = 'test',
    include_package_data=True,
    packages            = [
        'lazperf',
    ],
    classifiers         = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: GIS',
    ],
)
setup(ext_modules=extensions, **setup_args)

